#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Supplementary functions that are used in multiple classes
"""
import numpy as np
import scipy as sp
import sklearn.gaussian_process.kernels as kernels
from sklearn.gaussian_process import GaussianProcessRegressor
from scipy import linalg

# -------------------------------------------------------------------------
def check_ranges(theta, ranges):
    """
    This function checks if theta lies in the given ranges.

    Parameters
    ----------
    theta : array
        Proposed parameter set.
    ranges : nested list
        The parameter ranges.

    Returns
    -------
    c : bool
        If it lies in the given range, it returns True else False.

    """
    c = True
    # traverse in the list1
    for i, bounds in enumerate(ranges):
        x = theta[i]
        # condition check
        if x < bounds[0] or x > bounds[1]:
            c = False
            return c
    return c

# -------------------------------------------------------------------------
def hellinger_distance(p_, q_):
    """
    Hellinger distance between two continuous distributions.

    The maximum distance 1 is achieved when p_ assigns probability zero to
    every set to which q_ assigns a positive probability, and vice versa.
    0 (identical) and 1 (maximally different)

    Parameters
    ----------
    p_ : array
        Reference likelihood.
    q_ : array
        Estimated likelihood.

    Returns
    -------
    float
        Hellinger distance of two distributions.

    """
    p_ = np.array(p_)
    q_ = np.array(q_)

    mu1 = p_.mean()
    sigma1 = np.std(p_)

    mu2 = q_.mean()
    sigma2 = np.std(q_)

    term1 = np.sqrt(2 * sigma1 * sigma2 / (sigma1 ** 2 + sigma2 ** 2))

    term2 = np.exp(-.25 * (mu1 - mu2) ** 2 / (sigma1 ** 2 + sigma2 ** 2))

    h_squared = 1 - term1 * term2

    return np.sqrt(h_squared)


def logpdf(x, mean, cov):
    """
    Computes the likelihood based on a multivariate normal distribution.

    Parameters
    ----------
    x : TYPE
        DESCRIPTION.
    mean : array_like
        Observation data.
    cov : 2d array
        Covariance matrix of the distribution.

    Returns
    -------
    log_lik : float
        Log likelihood.

    """
    n = len(mean)
    l = linalg.cholesky(cov, lower=True)
    beta = np.sum(np.log(np.diag(l)))
    dev = x - mean
    alpha = dev.dot(linalg.cho_solve((l, True), dev))
    log_lik = -0.5 * alpha - beta - n / 2. * np.log(2 * np.pi)

    return log_lik


def subdomain(bounds, n_new_samples):
    """
    Divides a domain defined by Bounds into subdomains.

    Parameters
    ----------
    Bounds : list of tuples
        List of lower and upper bounds.
    n_new_samples : int
        Number of samples to divide the domain for.

    Returns
    -------
    Subdomains : List of tuples of tuples
        Each tuple of tuples divides one set of bounds into n_new_samples parts.

    """
    n_params = len(bounds)
    n_subdomains = n_new_samples + 1
    lin_space = np.zeros((n_params, n_subdomains))

    for i in range(n_params):
        lin_space[i] = np.linspace(start=bounds[i][0], stop=bounds[i][1],
                                   num=n_subdomains)
    subdomains = []
    for k in range(n_subdomains - 1):
        mylist = []
        for i in range(n_params):
            mylist.append((lin_space[i, k + 0], lin_space[i, k + 1]))
        subdomains.append(tuple(mylist))

    return subdomains


# noinspection SpellCheckingInspection
def corr_loocv_error(clf, psi, coeffs, y):
    """
    Calculates the corrected LOO error for regression on regressor
    matrix `psi` that generated the coefficients based on [1] and [2].

    [1] Blatman, G., 2009. Adaptive sparse polynomial chaos expansions for
        uncertainty propagation and sensitivity analysis (Doctoral
        dissertation, Clermont-Ferrand 2).

    [2] Blatman, G. and Sudret, B., 2011. Adaptive sparse polynomial chaos
        expansion based on least angle regression. Journal of computational
        Physics, 230(6), pp.2345-2367.

    Parameters
    ----------
    clf : object
        Fitted estimator.
    psi : array of shape (n_samples, n_features)
        The multivariate orthogonal polynomials (regressor).
    coeffs : array-like of shape (n_features,)
        Estimated cofficients.
    y : array of shape (n_samples,)
        Target values.

    Returns
    -------
    R_2 : float
        LOOCV Validation score (1-LOOCV erro).
    residual : array of shape (n_samples,)
        Residual values (y - predicted targets).

    """
    psi = np.array(psi, dtype=float)

    # Create PSI_Sparse by removing redundent terms
    nnz_idx = np.nonzero(coeffs)[0]
    if len(nnz_idx) == 0:
        nnz_idx = [0]
    psi_sparse = psi[:, nnz_idx]

    # NrCoeffs of aPCEs
    p_ = len(nnz_idx)
    # NrEvaluation (Size of experimental design)
    n_ = psi.shape[0]

    # Build the projection matrix
    psi_t_psi = np.dot(psi_sparse.T, psi_sparse)

    if np.linalg.cond(psi_t_psi) > 1e-12:  # and \
        # np.linalg.cond(PsiTPsi) < 1/sys.float_info.epsilon:
        # faster
        try:
            m_ = sp.linalg.solve(psi_t_psi,
                                 sp.sparse.eye(psi_t_psi.shape[0]).toarray())
        except:
            raise AttributeError(
                'There are too few samples for the corrected loo-cv error. '
                'Fit surrogate on at least as many '
                'samples as parameters to use this')
    else:
        # stabler
        m_ = np.linalg.pinv(psi_t_psi)

    # h factor (the full matrix is not calculated explicitly,
    # only the trace is, to save memory)
    psi_m = np.dot(psi_sparse, m_)

    h = np.sum(np.multiply(psi_m, psi_sparse), axis=1,
               dtype=np.longdouble)  # float128)

    # ------ Calculate Error Loocv for each measurement point ----
    # Residuals
    try:
        residual = clf.predict(psi) - y
    except:
        residual = np.dot(psi, coeffs) - y

    # Variance
    var_y = np.var(y)

    if var_y == 0:
        # norm_emp_error = 0
        loo_error = 0
        LCerror = np.zeros(y.shape)
        return 1 - loo_error, LCerror

    # norm_emp_error = np.mean(residual ** 2) / var_y

    # LCerror = np.divide(residual, (1-h))
    LCerror = residual / (1 - h)
    loo_error = np.mean(np.square(LCerror)) / var_y
    # if there are NaNs, just return an infinite LOO error (this
    # happens, e.g., when a strongly underdetermined problem is solved)
    if np.isnan(loo_error):
        loo_error = np.inf

    # Corrected Error for over-determined system
    tr_m = np.trace(m_)
    if tr_m < 0 or abs(tr_m) > 1e6:
        tr_m = np.trace(np.linalg.pinv(np.dot(psi.T, psi)))

    # Over-determined system of Equation
    if n_ > p_:
        T_factor = n_ / (n_ - p_) * (1 + tr_m)

    # Under-determined system of Equation
    else:
        T_factor = np.inf

    corrected_loo_error = loo_error * T_factor

    r_2 = 1 - corrected_loo_error

    return r_2, LCerror


def create_psi(basis_indices, univ_p_val):
    """
    This function assemble the design matrix Psi from the given basis index
    set INDICES and the univariate polynomial evaluations univ_p_val.

    Parameters
    ----------
    basis_indices : array of shape (n_terms, n_params)
        Multi-indices of multivariate polynomials.
    univ_p_val : array of (n_samples, n_params, n_max+1)
        All univariate regressors up to `n_max`.

    Raises
    ------
    ValueError
        n_terms in arguments do not match.

    Returns
    -------
    psi : array of shape (n_samples, n_terms)
        Multivariate regressors.

    """
    # Check if BasisIndices is a sparse matrix
    sparsity = sp.sparse.issparse(basis_indices)
    if sparsity:
        basis_indices = basis_indices.toarray()

    # Initialization and consistency checks
    # number of input variables
    n_params = univ_p_val.shape[1]

    # Size of the experimental design
    n_samples = univ_p_val.shape[0]

    # number of basis terms
    n_terms = basis_indices.shape[0]

    # check that the variables have consistent sizes
    if n_params != basis_indices.shape[1]:
        raise ValueError(
            f"The shapes of basis_indices ({basis_indices.shape[1]}) and "
            f"univ_p_val ({n_params}) don't match!!"
        )

    # Preallocate the Psi matrix for performance
    psi = np.ones((n_samples, n_terms))
    # Assemble the Psi matrix
    for m in range(basis_indices.shape[1]):
        aa = np.where(basis_indices[:, m] > 0)[0]
        try:
            basisIdx = basis_indices[aa, m]
            bb = univ_p_val[:, m, basisIdx].reshape(psi[:, aa].shape)
            psi[:, aa] = np.multiply(psi[:, aa], bb)
        except ValueError as err:
            raise err
    return psi


# -------------------------------------------------------------------------
def gaussian_process_emulator(X, y, nug_term=None, auto_select=False,
                              var_idx=None):
    """
    Fits a Gaussian Process Emulator to the target given the training
     points.

    Parameters
    ----------
    X : array of shape (n_samples, n_params)
        Training points.
    y : array of shape (n_samples,)
        Target values.
    nug_term : float, optional
        Nugget term. The default is None, i.e. variance of y.
    auto_select : bool, optional
        Loop over some kernels and select the best. The default is False.
    var_idx : int, optional
        The index number. The default is None.

    Returns
    -------
    gp : object
        Fitted estimator.

    """

    nug_term = nug_term if nug_term else np.var(y)

    kernels_ = [nug_term * kernels.RBF(length_scale=1.0,
                                       length_scale_bounds=(1e-25, 1e15)),
                nug_term * kernels.RationalQuadratic(length_scale=0.2,
                                                     alpha=1.0),
                nug_term * kernels.Matern(length_scale=1.0,
                                          length_scale_bounds=(1e-15, 1e5),
                                          nu=1.5)]

    # Automatic selection of the kernel
    if auto_select:
        gp = {}
        BME = []
        for i, kernel in enumerate(kernels_):
            gp[i] = GaussianProcessRegressor(kernel=kernel,
                                             n_restarts_optimizer=3,
                                             normalize_y=False)

            # Fit to data using Maximum Likelihood Estimation
            gp[i].fit(X, y)

            # Store the MLE as BME score
            BME.append(gp[i].log_marginal_likelihood())

        gp = gp[np.argmax(BME)]

    else:
        gp = GaussianProcessRegressor(kernel=kernels_[0],
                                      n_restarts_optimizer=3,
                                      normalize_y=False)
        gp.fit(X, y)

    # Compute score
    if var_idx is not None:
        score = gp.score(X, y)
        print('-' * 50)
        print(f'Output variable {var_idx}:')
        print('The estimation of GPE coefficients converged,')
        print(f'with the R^2 score: {score:.3f}')
        print('-' * 50)

    return gp
